/*
    IpToCountry 1.2
    Copyright (C) 2006 [es]Rush

    This program is free software; you can redistribute and/or modify
    it under the terms of the Open Unreal Mod License version 1.1.
*/

class LinkActor extends UBrowserHTTPClient config(IpToCountry);

var int ServersNum;

var config bool bSpawnAddon;
var config string ConfQueryServers[2];
var config int MaxTimeout;
var config int ErrorLimit;

struct QueryServer {
	var string IP;
	var string Host;
	var string URI;
	var int Port;
};
var QueryServer QueryServers[2];
var int Current;
var int Errors;

var string GMT; // GMT time that was sent by the query server

var bool bQueryInProgress;

var string QueryQueue[50];

var config string IPData[256];
var config int index; /* if IPData[] if full, the next data will be saved to IPData[index] */

var string QueryString;

var bool bDisabled;
var bool bCheckAOL;

var bool bContinueAtClose;
var bool bSwitchAtClose;

event PreBeginPlay()
{
	Tag='IpToCountry';
	Super.PreBeginPlay();
}

function PostBeginPlay()
{
	local Addon Ad;
	local int i;
	local string temp;
	local int SrcPort;

	if(ArrayCount(ConfQueryServers) != ArrayCount(QueryServers)) // just to point to you folks that this 2 arrays has to be the same length :>
	{
		log("IpToCountry - CRITICAL ERROR - BAD COMPILATION");
		Destroy();
	}
	else
		ServersNum=ArrayCount(ConfQueryServers);

	Super.PostBeginPlay();

	SaveConfig();

	if(bSpawnAddon)
	{
		Ad=Level.Spawn(class'Addon');
		Level.Game.BaseMutator.AddMutator(Ad);
		Ad.IpToCountry=self;
	}

	if(InStr(Caps(ConsoleCommand("get ini:Engine.Engine.GameEngine ServerPackages")), Caps("IpToCountry_AOL")) != -1)
		bCheckAOL=True;

	log("+----");
	log("| IpToCountry 1.2 by [es]Rush - Enabled!");

	for(i=0;i<ArrayCount(ConfQueryServers);i++)
	{
		if(InStr(ConfQueryServers[i], ",") != -1)
		{
			QueryServers[i].IP=FixIPAddr(SepLeft(ConfQueryServers[i], ","));
			temp=SepRight(ConfQueryServers[i], ",");
			QueryServers[i].Host=SepLeft(temp, "/");

			SrcPort=int(SepRight(QueryServers[i].Host, ":"));
			if(SrcPort == 0)
				SrcPort=80;
			QueryServers[i].Port=SrcPort;
			QueryServers[i].Host=SepLeft(QueryServers[i].Host, ":");
			QueryServers[i].URI=SepRight(temp, "/");
			temp=SepLeft(ConfQueryServers[i], ","); // store the bad ip address to print it on error
		}
		else
		{
			QueryServers[i].Host=SepLeft(ConfQueryServers[i], "/");

			SrcPort=int(SepRight(QueryServers[i].Host, ":"));
			if(SrcPort == 0)
				SrcPort=80;
			QueryServers[i].Port=SrcPort;
			QueryServers[i].Host=SepLeft(QueryServers[i].Host, ":");
			QueryServers[i].URI=SepRight(ConfQueryServers[i], "/");

			QueryServers[i].IP=FixIpAddr(QueryServers[i].Host);
			temp=QueryServers[i].Host; // store the bad ip address to print it on error
		}

		if(QueryServers[i].IP == "")
		{
			log("| QueryServers["$i$"]   : IP '"$temp$"' is unusable");
			log("| IpToCountry is disabled! Fix the problem in IpToCountry.ini");
			log("+----");
			bDisabled=True;
			return;
		}


		if(QueryServers[i].IP == "")
		{
			if(InStr(ConfQueryServers[i], ",") != -1)
				log("| QueryServers["$i$"]   : IP '"$SepLeft(ConfQueryServers[i], ",")$"' is unusable");
			else // ok, let's try to
			{
				QueryServers[i].Host=SepLeft(temp, "/");


				if(QueryServers[i].IP == "")
					log("| QueryServers["$i$"]   : IP address of "$SepLeft(ConfQueryServers[i], "/")$" hasn't been specified.");
				else
				{
					SrcPort=int(SepRight(QueryServers[i].Host, ":"));
					if(SrcPort == 0)
						SrcPort=80;
					QueryServers[i].Port=SrcPort;
				}
			}

			if(QueryServers[i].IP == "") // if the problem still remains ...
			{
				log("| IpToCountry is disabled! Fix the problem in IpToCountry.ini");
				log("+----");
				bDisabled=True;
				return;
			}
		}
		if(QueryServers[i].Port == 80)
			log("| QueryServers["$i$"]   : "$QueryServers[i].Host$"/"$QueryServers[i].URI);
		else
			log("| QueryServers["$i$"]   : "$QueryServers[i].Host$":"$QueryServers[i].Port$"/"$QueryServers[i].URI);
	}
	log("| Extension for AOL : "$bCheckAOL);
	log("+----");


	// testing
	for(i=0;i<ArrayCount(IPData);i++)
	{
		IPData[i]="";
	}
	SaveConfig();
	AddToQueue("153.19.20.10");
	AddToQueue("153.19.48.15");
	//
}

function string GetItemName(string IP)
{
	local int i;

	/* a way of TimeReplication to communicate with this actor */
	if(Left(IP, 7) == "AOLSUX ")
		AOLHandler(Mid(IP, 7));

	if(bDisabled)
		return "!Disabled";

	IP=FixIpAddr(IP);
	if(IP == "")
		return "!Bad Input";

	i=IpInDatabase(IP);
	if(i!=-1)
	{
		if(InStr(IPData[i], "AMERICA ONLINE") == -1)
			return IPData[i];
		else
			return "!AOL - Trying to clarify";
	}
	else if(IpInQueue(IP))
		return "!Waiting in queue";
	else if(InStr(QueryString, IP) != -1)
		return "!Resolving now";
	else
	{
		if(QueryQueue[49] != "")
			return "!Queue full";
		else
		{
			AddToQueue(IP);
			SendQueue();
			return "!Added to queue";
		}
	}
}

function HTTPReceivedData (string data)
{
	local string result, temp;
	local int elems, i;
	local string IP;

	result = ParseString(data);
	elems = ElementsNum(result, ",");

	Super.SetTimer(0.0, false); // disable the timeout count

	GMT=left(result, 5); // get the GMT time of a query server
	result=Mid(result, 6); // leave the rest of the string in the good old format

	if(InStr(result, "Warning") != -1) // php warning ?
	{
		SwitchQueryServer("IpToCountry: "$QueryServers[Current].Host$" returned bad data! Switching to alternate.");
		SendQueue();
		return;
	}
	if(elems==1)
	{
		IP=SelElem(result, 1);
		if(ElementsNum(result)!=5  || FixIpAddr(IP) == "") // some weird data ?
		{
			SwitchQueryServer("IpToCountry: "$QueryServers[Current].Host$" returned bad data! Switching to alternate.");
			SendQueue();
			return;
		}
		if(Right(SelElem(result, 2), 8) == ".aol.com" && bCheckAOL)
		{
			SaveIPData(SetAOL(result)); /* save IP to keep the record, this data won't be returned anywhere anyway */
			InitAOLCheck(IP);
		}
		else
   		SaveIPData(result);
   }
 	else
 	{
 	 	for(i=1;i<=elems;i++)
 	 	{
 	 		temp=SelElem(result, i, ",");
 	 		IP=SelElem(temp, 1);
			if(i==1)
			{
				if(ElementsNum(temp)!=5 || FixIpAddr(IP) == "") // some weird data ?
				{
					SwitchQueryServer("IpToCountry: "$QueryServers[Current].Host$" returned bad data! Switching to alternate.");
					SendQueue();
					return;
				}
			}
			if(Right(SelElem(temp, 2), 8) == ".aol.com" && bCheckAOL)
			{
				SaveIPData(SetAOL(temp));
				InitAOLCheck(IP);
			}
			else
				SaveIPData(temp);
 	 	}
 	}
 	QueryString="";
	bQueryInProgress=False;
	SendQueue();
}

function bool IpOnServer(string IP)
{
	local Pawn P;

	for( P=Level.PawnList; P!=None; P=P.NextPawn )
		if(P.IsA('PlayerPawn'))
			if(NetConnection(PlayerPawn(P).Player) != None)
				if(SepLeft(PlayerPawn(P).GetPlayerNetworkAddress()) == IP)
					return true;
	return false;
}

function string SetAOL(string Input)
{
	return SelElem(Input, 1)$":"$SelElem(Input, 2)$":AMERICA ONLINE:AOL:us";
}

function bool InitAOLCheck(string IP)
{
	local Pawn P;
	local TimeReplication TR;

	for( P=Level.PawnList; P!=None; P=P.NextPawn )
		if(P.IsA('PlayerPawn'))
			if(NetConnection(PlayerPawn(P).Player) != None)
			{
				if(SepLeft(PlayerPawn(P).GetPlayerNetworkAddress()) == IP)
				{
					TR=Spawn(class'TimeReplication', P);
					TR.IpToCountry=self;
					return true;
				}
			}
	return false;
}

event Opened()
{
	Enable('Tick');
	if(ProxyServerAddress != "")
		SendBufferedData("GET http://"$ServerAddress$":"$string(ServerPort)$ServerURI$" HTTP/1.1"$CR$LF);
	else
		SendBufferedData("GET "$ServerURI$" HTTP/1.1"$CR$LF);
	SendBufferedData("User-Agent: Unreal"$CR$LF);
	SendBufferedData("Connection: close"$CR$LF);
	SendBufferedData("Host: "$QueryServers[Current].Host$":"$ServerPort$CR$LF$CR$LF);

	CurrentState = WaitingForHeader;
}

function string AOLHandler(string Data)
{
	local int ClientTime;
	local string IP;

	IP=SelElem(Data, 1);
	ClientTime = int(SelElem(Data, 2))*60 + int(SelElem(Data, 3));

	UpdateAOLData(IP$":"$GetAOLArrayCountry(ClientTime));
}

function string GetAOLArrayCountry(int ClientTime)
{
   local int GMTTime;
   local int TimeDiff;
   local string ArrayCountry;

   GMTTime = int(SelElem(GMT, 1))*60 + int(SelElem(GMT, 1));
   TimeDiff = abs(GMTTime - ClientTime) % 1440;
   if (TimeDiff > 720) TimeDiff = abs(TimeDiff - 1440);

   if (TimeDiff < 30)
   	ArrayCountry = "UNITED KINGDOM:GBR:gb";
   else if (TimeDiff < 120)
   	ArrayCountry = "GERMANY:DEU:de";
   else
   	ArrayCountry = "UNITED STATES:USA:us";
   return ArrayCountry;
}


function SwitchQueryServer(optional string LogStr)
{
	if(LogStr != "")
		log(LogStr);
	bQueryInProgress=False;

	if(++Errors > ErrorLimit)
	{
		log("IpToCountry: Too many errors in one session, IpToCountry disabled.");
		bDisabled=True;
	}
	Current=(Current+1) % ServersNum;

	ServerIpAddr.Addr=0; // ensures that the Browse() function will open a connection to a new address
}

// override original function to silent the warning log
function DoBind()
{
	if( BindPort() == 0 )
	{
		SetError(-2);
		return;
	}
	Open( ServerIpAddr );
	bClosed = False;
}

event ResolveFailed()
{
	SetTimer(0.0, false);
	SetError(-4);
}

function SetError(int code)
{
	Super.SetError(code);

	switch(code)
	{
		case -4:
			log("IpToCountry: Error at resolving IP: "$QueryServers[Current].IP$"!");
			break;
		case -1:
			log("IpToCountry: Error at resolving host: "$QueryServers[Current].Host$"!");
			break;
		case -2:
			log( "IpToCountry: Error binding local port." );
			break;
		case -3:
			log("IpToCountry: "$QueryServers[Current].Host$" timed out after "$string(MaxTimeout)$" seconds!");
			break;
		default:
			log("IpToCountry: Got HTTP error with code "$string(code)$" from "$QueryServers[Current].Host$"!");
	}
	 // sometimes the connection doesn't break immediately, it is probably due to some bug in UBrowserHTTPClient, if it happens we have to wait for it inside event Closed() cause we cannot open the same socket if it is already opened
	if(IsConnected())
		bSwitchAtClose=True;
	else
	{
		SwitchQueryServer("IpToCountry: "$QueryServers[Current].Host$" failed! Trying the alternate.");
		SendQueue();
	}
}

event Closed()
{
	Super.Closed();
	bQueryInProgress=False;
	if(bSwitchAtClose)
	{
		bSwitchAtClose=False;
		SwitchQueryServer("IpToCountry: "$QueryServers[Current].Host$" failed! Trying the alternate.");
		SendQueue();
	}
	else if(bContinueAtClose)
	{
		bContinueAtClose=False;
		bQueryInProgress=False;
		SendQueue();
	}
}

function SendQueue()
{
	local int i;
	local string IP;
	local bool bNotFirst;

	if(bDisabled)
		return;
	if(IsConnected())
		bContinueAtClose=True;

	if(bQueryInProgress || (QueryQueue[0] == "" && QueryString == ""))
		return;

	for(i=0;i<ArrayCount(QueryQueue);i++)
	{
		if(QueryQueue[i] == "")
			continue;
		IP=QueryQueue[i];
		if(QueryString=="")
			QueryString=QueryQueue[i];
		else
			QueryString=QueryString$","$QueryQueue[i];
		QueryQueue[i]="";
	}
	if(QueryString == "")
		return;

	bQueryInProgress=True;

	Browse(QueryServers[Current].IP,"/"$QueryServers[Current].URI$"?ip="$QueryString,QueryServers[Current].Port,MaxTimeout);
}

function AddToQueue(string IP)
{
	local int i;

	if(IpInDatabase(IP) != -1 || InStr(QueryString, IP) != -1)
	{
//		log("IpToCountry(DEBUG): AddToQueue("$IP$") - but IP is already in the queue");
		return;
	}
//	else
//		log("IpToCountry(DEBUG): AddToQueue("$IP$")");
	for(i=0;i<ArrayCount(QueryQueue);i++)
	{
		if(QueryQueue[i]!="")
			continue;
		QueryQueue[i] = IP;
		if(!bQueryInProgress)
			SendQueue();
		break;
	}
}

function UpdateAOLData(string SetString)
{
	local int i;
	local string IP, Rest;

	IP=SelElem(SetString, 1);
	Rest=Mid(SetString, InStr(SetString, ":")+1); /* rest of the elements */

	for(i=0;i<ArrayCount(IPData);i++)
	{
		if(SepLeft(IPData[i]) == IP)
		{
			IPData[i] = IP$":"$SelElem(IPData[i], 2)$":"$Rest;
			SaveConfig();
			return;
		}
	}
	SaveIPData(SetString); /* just in case there's nothing to alter, add new data */
}

function bool SaveIPData(string SetString)
{
	local int i;

//	log("IpToCountry(DEBUG): Adding IP data "$SetString);

	for(i=0;i<ArrayCount(IPData);i++)
	{
		if(IPData[i] != "")
			continue;
		IPData[i]=SetString;
		SaveConfig();
		return true;
	}
	IPData[index]=SetString;
	index = (index+1) % ArrayCount(IPData);
	SaveConfig();
}

function int IpInDatabase(string IP)
{
	local int i;

	for(i=0;i<ArrayCount(IPData);i++)
	{
		if(SepLeft(IPData[i]) == IP)
			return i;
	}
	return -1;
}

function bool IpInQueue(string IP)
{
	local int i;
	local string host;

	for(i=0;i<ArrayCount(QueryQueue);i++)
	{
		if(QueryQueue[i] == IP)
			return true;
	}
	return false;
}

/* STRING FUNCTIONS */

// Arguments(Input string, Character separating elements)
static final function int ElementsNum(string Str, optional string Char)
{
	local int count, pos;

	if(Char=="")
		Char=":";

	while(true)
	{
		pos = InStr(Str, Char);
		if(pos == -1)
			break;
		Str=Mid(Str, pos+1);
		count++;
	}
	return count+1;
}

// Arguments(Input string, Element to get, Character separating elements)
static final function string SelElem(string Str, int Elem, optional string Char)
{
	local int pos;
	if(Char=="")
		Char=":";

	while(Elem>1)
	{
		Str=Mid(Str, InStr(Str, Char)+1);
		Elem--;
	}
	pos=InStr(Str, Char);
	if(pos != -1)
    	Str=Left(Str, pos);
    return Str;
}

static final function string SepLeft(string Input, optional string Char)
{
	local int pos;
	if(Char=="")
		Char=":";

	pos = InStr(Input, Char);
	if(pos != -1)
		return Left(Input, pos);
	else
		return Input;
}

static final function string SepRight(string Input, optional string Char)
{
	local int pos;
	if(Char=="")
		Char=":";

	pos = InStr(Input, Char);
	if(pos != -1)
		return Right(Input, len(Input)-pos-1);
	else
		return "";
}

function string ParseString (String Input)
{
	local int CRLF,i;
	local string result, temp;
	local bool b;

	CRLF = InStr(Input ,CR$LF);

	// No CR or LF in string
	if (CRLF == -1)
		return Input;
	else
	{
		result = Right(Input, len(Input)-CRLF-2);
		CRLF = InStr(result ,CR$LF);
		result = Left(result, CRLF);
		return result;
    }
}

function string FixIpAddr(string IP)
{
	local IpAddr Addr;
	local int i;
	local String StrAddr;

	if(!StringToIpAddr(IP, Addr))
		return "";
	StrAddr=IpAddrToString(Addr);
	i = InStr(StrAddr, ":");
	if(i != -1)
		StrAddr = Left(StrAddr, i);
	else
		return "";
	return StrAddr;
}

defaultproperties
{
	Tag='IpToCountry'
	bSpawnAddon=True;
	ConfQueryServers(0)="85.236.100.16,utgl.unrealadmin.org/iptocountry12.php"
	ConfQueryServers(1)="153.19.48.14,153.19.48.14/iptocountry12.php"
	MaxTimeout=10
	ErrorLimit=10
}